#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>

#define HIGHEST_BIT    sizeof( unsigned int ) * 8 - 1 /* faulheit ;-) */

#define ABNORMAL_TERM           -1  /* exitcode bei fehler */
/* kann noch verfeinert werden: verschiedene exitcodes
 fuer verschiedene Fehler */
#define MASTER                  0  /* sollte immer auf 0 definiert sein, da es */
/* sonst probleme mit der interaktion gibt
 (siehe 1. Versuch MPI) */
#define TAG                     17  /* universell eingesetztes TAG */
#define TRUE                    1
#define FALSE                   0
#define MAX_VALUE               256.0 /* maximaler wert, fuer die zufaellig erzeugten */
/* Daten.*/
#define CART_PROC_PER_DIMENSION 2  /* anz. prozesse pro dim. im cart. hypercube */
/* (siehe definition hypercube und parameterliste
  von MPI_Cart_create() ) */
#define CART_PERIOD_DIMENSION   1  /* legt fest, ob dimensionen zyklisch sind oder nicht */
/* (siehe definition hypercube und parameterliste
  von MPI_Cart_create() ) */
#define REORDER     1  /* Parameter fuer MPI_Cart_create() / MPI_Graph_create() */
#define NO_REORDER    0

/*
 * speichert ein globales minimum und den rank des prozesses,
 * bei dem dieses gefunden wurde
 */
struct __val_rank_pair {
	int val;
	int rank;
};
typedef struct __val_rank_pair val_rank_pair;

/*
 Erzeugt einen Hypercube vom angegebenen Communicator srcComm und
 liefert fuer den Hypercube den  Communicator dstComm zurueck.
 Diese Funktion sollte beim Erzeugen der Communicatoren als erste
 aufgerufen werden, da sie auch die Dimension und Anzahl der Knoten
 im Hypercube zurueckgibt. Diese koennen dann als Eingabeparameter der
 anderen Erzeugungsfunktionen benutzt werden.

 Parameter:
 srcComm - IN,  Quellcommunicator
 dstComm - OUT, Zielcommunicator
 dim     - OUT, Dimension des Hypercubes
 size    - OUT, Anzahl der Knoten im Hypercube

 Return:
 ABNORMAL_TERM, wenn ein Fehler passiert. Mit diesem Wert kann dann das
 Programm auch verlassen werden.
 0 sonst
*/
int make_hyper(MPI_Comm srcComm, MPI_Comm *dstComm, int *dim, int *size);

/*
 Fuehrt auf dem angegebnen Communicator einen hypercubeartigen
 Broadcast aus. Der hier mit master bezeichnete Knoten bezieht
 sich auf den uebergebenen Communicator und stellt die
 Datenquelle dar.

 Parameter:
 dim    - IN,  Dimension des Hypercubes
 mister - IN,  master, von dem aus die Communication beginnt
 cube   - IN,  Communicator, auf dem der Broadcast stattfindet
 buf    - IN,  zu versendender Buffer
 count  - IN,  Anzahl der Bytes in buf
 type   - IN,  MPI_Datentyp der Elemente in buf
 tag    - IN,  TAG fuer Kommunikation
 status - OUT, Status der Kommunikation
*/
void hyper_bcast(int dim, int master, MPI_Comm cube,
                 void *buf, int count,
                 MPI_Datatype type, int tag,
                 MPI_Status *status);

/*
 Liest die Anzahl der Daten und die Daten selbst ein.
 Verteilt die Daten auf die einzelnen Knoten, die sich
 im angegebenen Communicator befinden.
 Das Einlesen kann nur vom globalen Master erfolgen. Dieser
 wird hier nochmal uebergeben, genau wie der Welt-Communicator,
 da beide benutzt werden koennen, um auf den Rank des Welt-MAsters
 im Communicator cube zu schliessen. Dies ist ein wichtiger Schritt
 in der Funktion.

 Parameter:
 master_in_world - IN,  Rank des Masters in der Welt
 world_comm      - IN,  Welt-Communicator
 dim             - IN,  Dimension des Hypercubes
 tag             - IN,  TAG fuer das Verteilen der Daten
 cube            - IN,  Communicator, auf dem die Daten verteilt
         werden sollen
 status          - OUT, Status der Kommunikation
 count           - OUT, Anzahl der Daten, die der Prozess erhalten hat
 data            - OUT, Array, in dem sich die Daten befinden.
*/
int hyper_init_data(int master_in_world, MPI_Comm world_comm,
                    int dim, int tag,
                    MPI_Comm cube, MPI_Status *status,
                    int *count, int **data);

/*
 Berechnet die globale Summe der auf die Knoten verteilten Daten.

 Parameter:
 buffer - IN,  Buffer ueber den Summiert werden soll
 len    - IN,  Laenge des Buffers (Anzahl Elemente)
 dim    - IN,  Dimension des Hypercubes
 cube   - IN,  Communicator, auf dem summiert werden soll
 sum    - OUT, globale Summe.
*/
void hyper_global_sum(int *buffer, int len, int dim, MPI_Comm cube, int *sum);

/*
 Berechnet das globale Minimum der Daten, die auf die Knoten
 des uebergebenen Communicators verteilt sind

 Parameter:
 buffer - IN,  Buffer, in dem das Minimum gesucht werden soll
 len    - IN,  Laenge des Buffers (Anzahl Elemente)
 dim    - IN,  Dimension des Hypercubes
 cube   - IN,  Communicator, auf dem das globale Minimum gesucht wird
 vrp    - OUT, Wert-Rank-Paar, das das globale Minimum sowie dessen
               Position enthaelt
*/
void hyper_global_min(int *buffer, int len, int dim, MPI_Comm cube, val_rank_pair *vrp);


/*
 * funktionen fuer die verwaltung des cartesischen topologie - cubes
 */


/*
 Erzeugt einen Hypercube mit Hilfe der cartesischen Topologie und
 liefert den Communicator mit der max. Anzahl Prozesse zurueck

 Parameter:
 srcComm - IN,  QuellCommunicator
 dstComm - OUT, ZielCommunicator
 dim     - IN,  Dimension des Hypercubes
*/
int cart_make_hyper(MPI_Comm srcComm, MPI_Comm *dstComm, int ndim);

/*
 Liefert den Nachbar des angegebenen Knotens in der angegebenen
 Dimension. Diese Operation bezieht sich auf den angegebenen
 Communicator.

 Parameter:
 cube      - IN,  Communicator des Hypercubes, in dem der Nachbar gesucht werden soll
 maxdim    - IN,  max. Dimension des Hypercubes
 dim       - IN,  Dimension, in der Nachbar bestimmt werden soll
 rank      - IN,  der Rank, zu dem der Nachbar bestimmt werden soll
 otherRank - OUT, der Rank des Nachbarn
*/
int cart_hyper_get_neighbour(MPI_Comm cube, int maxdim, int dim, int rank, int *otherRank);

/*
 Berechnet die globale Summe der auf die Knoten verteilten Daten.

 Parameter:
 buffer - IN,  Buffer ueber den summiert werden soll
 len    - IN,  Laenge des Buffers (Anzahl Elemente)
 dim    - IN,  Dimension des Hypercubes
 cube   - IN,  Communicator, auf dem summiert werden soll
 sum    - OUT, globale Summe.
*/
void cart_hyper_global_sum(int *buffer, int len, int dim, MPI_Comm cube, int *sum);

/*
 Berechnet das globale Minimum der Daten, die auf die Knoten
 des uebergebenen Communicators verteilt sind

 Parameter:
 buffer - IN,  Buffer, in dem das Minimum gesucht werden soll
 len    - IN,  Laenge des Buffers (Anzahl Elemente)
 dim    - IN,  Dimension des Hypercubes
 cube   - IN,  Communicator, auf dem das globale Minimum gesucht wird
 vrp    - OUT, Wert-Rank-Paar, das das globale Minimum sowie dessen
               Position enthaelt
*/
void cart_hyper_global_min(int *buffer, int len, int dim, MPI_Comm cube, val_rank_pair *vrp);


/*
 * funktionen fuer die verwaltung des graphen topologie - cubes
 */


/*
 Erzeugt einen Hypercube mit Hilfe der Graphen-Topologie und
 liefert den Communicator mit der max. Anzahl Prozesse zurueck

 Parameter:
 srcComm - IN,  QuellCommunicator
 dstComm - OUT, ZielCommunicator
 dim     - IN,  Dimension des Hypercubes
*/
int graph_make_hyper(MPI_Comm srcComm, MPI_Comm *dstComm, int ndim);

/*
 Liefert den Nachbar des angegebenen Knotens in der angegebenen
 Dimension. Diese Operation bezieht sich auf den angegebenen
 Communicator.

 Parameter:
 cube      - IN,  Communicator des Hypercubes, in dem der Nachbar gesucht werden soll
 ndim      - IN,  max. Dimension des Hypercubes
 dim       - IN,  Dimension, in der Nachbar bestimmt werden soll
 rank      - IN,  der Rank, zu dem der Nachbar bestimmt werden soll
 otherRank - OUT, der Rank des Nachbarn
*/
int graph_hyper_get_neighbour(MPI_Comm cube, int ndim, int dim, int rank, int *otherRank);

/*
 Berechnet die globale Summe der auf die Knoten verteilten Daten.

 Parameter:
 buffer - IN,  Buffer ueber den summiert werden soll
 len    - IN,  Laenge des Buffers (Anzahl Elemente)
 dim    - IN,  Dimension des Hypercubes
 cube   - IN,  Communicator, auf dem summiert werden soll
 sum    - OUT, globale Summe.
*/
void graph_hyper_global_sum(int *buffer, int len, int dim, MPI_Comm cube, int *sum);

/*
 Berechnet das globale Minimum der Daten, die auf die Knoten
 des uebergebenen Communicators verteilt sind

 Parameter:
 buffer - IN,  Buffer, in dem das Minimum gesucht werden soll
 len    - IN,  Laenge des Buffers (Anzahl Elemente)
 dim    - IN,  Dimension des Hypercubes
 cube   - IN,  Communicator, auf dem das globale Minimum gesucht wird
 vrp    - OUT, Wert-Rank-Paar, das das globale Minimum sowie dessen
               Position enthaelt
*/
void graph_hyper_global_min(int *buffer, int len, int dim, MPI_Comm cube, val_rank_pair *vrp);
